""" Este script implementa una busqueda aleatoria (Random Search) para encontrar el 
    modelo con mejor precisión. 
    Concretamente, hemos parametrizado tres modelos: XGBRegressor, LGBMRegressor y 
    CatBoostRegressor. 
    Usamos las rebajas de verano de 2016 como validación y entrenamos con todo lo
    anterior.        
"""

import pandas as pd
import numpy as np

from datetime import datetime, timedelta
from tqdm import tqdm

from sklearn.model_selection import ParameterSampler
from scipy.stats import loguniform
from xgboost import XGBRegressor
from lightgbm import LGBMRegressor
from catboost import CatBoostRegressor
from sklearn.neighbors import KNeighborsRegressor
from sklearn.linear_model import LinearRegression
from sklearn.ensemble import (
    AdaBoostRegressor,
    BaggingRegressor,
    ExtraTreesRegressor,
    GradientBoostingRegressor,
    StackingRegressor,
    RandomForestRegressor,
)

from seleccion_modelo_utils import (
    get_model_inputs,
    get_error_metrics,
    read_dataset,
    normalizar_visitas,
    generate_features_timeseries,
    generate_date_features,
)

RANDOM_STATE = 314159

n_jobs = 8
n_iter = 100

try:
    results = pd.read_csv(".results.csv", index_col=0).to_dict(orient="index")
except:
    results = dict()

models = [
    (
        XGBRegressor,  # gbtree
        {
            "booster": ["gbtree"],
            "nthread": [n_jobs],
            "n_estimators": list(range(10, 501, 10)),  # 10, 20, ... 500
            "eta": loguniform(1e-4, 0.5),
            "min_child_weight": loguniform(1e-3, 1.0),
            "max_depth": list(range(3, 50)),
            "gamma": loguniform(1e-4, 2),
            "subsample": loguniform(0.1, 1),
            "colsample_bytree": loguniform(0.1, 1),
            "lambda": loguniform(1e-5, 5.0),
            "alpha": loguniform(1e-5, 5.0),
            "eval_metric": ["mae"],
        },
    ),
    (
        XGBRegressor,  # gblinear
        {
            "booster": ["gblinear"],
            "nthread": [n_jobs],
            "n_estimators": list(range(10, 501, 10)),
            "eta": loguniform(0.001, 0.5),
            "lambda": loguniform(1e-5, 5.0),
            "alpha": loguniform(1e-5, 5.0),
            "eval_metric": ["rmse", "mae"],
        },
    ),
    (
        CatBoostRegressor,
        {
            "thread_count": [n_jobs],
            "iterations": list(range(10, 501, 10)),
            "learning_rate": loguniform(1e-5, 0.5),
            "l2_leaf_reg": loguniform(1e-5, 5.0),
            "bootstrap_type": ["Bayesian"],
            "bagging_temperature": loguniform(1e-2, 10.0),
            "depth": list(range(2, 17)),
            "grow_policy": ["SymmetricTree", "Depthwise", "Lossguide"],
            "min_data_in_leaf": list(range(1, 10)),
            "loss_function": ["RMSE", "MAE", "MAPE"],
            "logging_level": ["Silent"],
        },
    ),
    (
        LGBMRegressor,
        {
            "objective": ["regression"],
            "n_jobs": [n_jobs],
            "boosting_type": ["gbdt", "dart", "rf"],
            "num_leaves": list(range(10, 151, 5)),
            "max_depth": [-1] + list(range(3, 20)),
            "learning_rate": loguniform(1e-5, 0.5),
            "n_estimators": list(range(10, 1001, 10)),  # 10, 20, ... 500
            "min_split_gain": loguniform(1e-7, 1.0),
            "min_child_weight": loguniform(1e-2, 5.0),
            "min_child_samples": list(range(2, 51)),
            "subsample": loguniform(0.25, 0.999),
            "subsample_freq": list(range(1, 101, 5)),
            "colsample_bytree": loguniform(0.25, 1),
            "reg_lambda": loguniform(1e-5, 5.0),
            "reg_alpha": loguniform(1e-5, 5.0),
            "metric": ["mae"],
        },
    ),
    (
        RandomForestRegressor,
        {
            "bootstrap": [True, False],
            "max_depth": [10, 20, 100],
            "max_features": ["auto", "sqrt"],
            "min_samples_leaf": [1, 2, 4],
            "min_samples_split": [2, 5, 10],
            "n_estimators": [9, 200, 1800, 2000],
        },
    ),
    (
        AdaBoostRegressor,
        {
            "n_estimators": [50, 70],
            "learning_rate": [0.01, 0.05, 0.1, 0.3, 1],
            "loss": ["linear", "square", "exponential"],
        },
    ),
    (
        BaggingRegressor,
        {
            "base_estimator": [None, LinearRegression(), KNeighborsRegressor()],
            "n_estimators": [20, 50, 100],
            "max_samples": [0.5, 1.0],
            "max_features": [0.5, 1.0],
            "bootstrap": [True, False],
            "bootstrap_features": [True, False],
        },
    ),
    (
        ExtraTreesRegressor,
        {
            "n_estimators": [50, 75, 100],
            "max_features": [50, 100],
            "min_samples_leaf": [20, 40],
            "min_samples_split": [15, 25],
        },
    ),
    (
        GradientBoostingRegressor,
        {
            "n_estimators": [500, 1000, 2000],
            "learning_rate": [0.001, 0.01, 0.1],
            "max_depth": [1, 2, 4],
            "subsample": [0.5, 0.75, 1],
            "random_state": [1],
        },
    ),
]


def run_experiment(
    model_fn, space_params, n_iter, X_train, y_train, X_test, y_test, results=dict()
):
    search = ParameterSampler(space_params, n_iter, random_state=RANDOM_STATE)
    for params in tqdm(search):
        model = model_fn()
        model.set_params(**params)
        model.fit(X_train, y_train)
        o = model.predict(X_test)
        err = get_error_metrics(y_test, o)
        results[f"{type(model).__name__} - {params}"] = err
        pd.DataFrame(results).T.sort_values("CUSTOM_METRIC").to_csv("./results.csv")

    return results


# Leer los datos
print("> Reading dataset")
df = read_dataset("data/Modelar_UH2021.txt", "data/Estimar2.txt")

# Preprocesar los datos: Normalizar y generar características.
print("> Preprocess dataset")
df = df[
    [
        "id",
        "fecha",
        "visitas",
        "unidades_vendidas",
        "precio",
        "antiguedad",
        "campaña",
        "dia_atipico",
        "train/test",
    ]
]
df = normalizar_visitas(df)
df = generate_features_timeseries(df, columns=["visitas"])
df = generate_date_features(df, keep_fecha=True)
df = df[(df.notna().all(axis=1)) | (df["train/test"] == "test")]

# Dividir en training/test
lim_fecha = datetime.strptime("2016-06-01", "%Y-%m-%d")
lim_fecha2 = datetime.strptime("2016-07-31", "%Y-%m-%d") + timedelta(days=7)

df_train = df[(df["fecha"] < lim_fecha)]
df_train = df_train[(df_train.notna().all(axis=1))]
X_train, y_train = get_model_inputs(df_train)

test = df[(df["fecha"] < lim_fecha2) & (df["fecha"] >= lim_fecha)]
test = test[(test.notna().all(axis=1))]
X_test, y_test = get_model_inputs(test)

print(X_train.shape, X_test.shape)


print("> Start experiments\n")
for i, (model_fn, model_params) in enumerate(models):
    print(i, model_fn.__name__)
    results = run_experiment(
        model_fn,
        model_params,
        n_iter,
        X_train,
        y_train,
        X_test,
        y_test,
        results=results,
    )
