import pandas as pd
import numpy as np
from datetime import datetime
from sklearn.preprocessing import StandardScaler


def read_dataset(
    train_path="data/Modelar_UH2021.txt",
    test_path="data/Estimar2.txt",
    target_feat="unidades_vendidas",
):
    '''
    Función de lectura del dataset

    :param train_path: Ruta al fichero con los datos de entrenamiento
    :param test_path: Ruta al fichero con los datos de test
    :param target_feat: Nombre de la columna de la variable a predecir
    :return:
        Dataframe con todos los datos procesados

    '''

    # Unimos los conjuntos de train y test en un único dataframe, con una columna 'train/test'
    df = pd.read_csv(train_path, sep="|")
    df["train/test"] = "train"
    df_test = pd.read_csv(test_path, sep="|")
    df_test["train/test"] = "test"
    df_test[target_feat] = np.NaN
    df = pd.concat([df, df_test], ignore_index=True, sort=False)
    del df_test

    # Casting de las columnas fecha y precio al formato correcto.
    df["fecha"] = pd.to_datetime(df["fecha"], dayfirst=True)
    df["precio"] = df["precio"].apply(lambda x: float(str(x).replace(",", ".")))

    # Fill valores nulos de precios
    df["precio"] = df.groupby(["id"])["precio"].ffill()
    df["precio"] = df.groupby(["id"])["precio"].bfill()

    # Fill valores nulos de categoria_dos
    df["categoria_dos"] = df["categoria_dos"].fillna(-1)

    # Fill valores nulos de antigüedad
    df["antiguedad"] = df["antiguedad"].apply(lambda x: 0 if x == "-" else x)
    df["antiguedad"] = df.groupby(["id"])["antiguedad"].bfill()
    df["antiguedad"] = df.groupby(["id"])["antiguedad"].ffill()
    df["antiguedad"] = df["antiguedad"].fillna(value=0)

    # Eliminar filas duplicadas
    df = df.drop_duplicates()
    # Eliminar filas duplicadas que tienen distinto valor de campaña (elegimos eliminar los 1 o los 0)
    df = df.drop(
        _get_indices_to_remove_dup_campana(df[df["train/test"] == "train"], campana=0)
    )

    return df



def _get_indices_to_remove_dup_campana(df, campana=1):
    '''

    :param df: dataframe completo
    :param campana: valor del atributo campaña que se desea eliminar
    :return:
        Índices del dataframe a eliminar para eliminar duplicados
    '''
    # Hay entradas duplicadas donde solo cambia el valor del atributo "campaña"
    # En esta función cogemos los índices de esos duplicados con valores 0 o 1 según el parámetro de entrada
    # para poder luego eliminarlos del dataframe
    df = df.drop_duplicates()
    df_fecha_id = df[["fecha", "id"]]
    df_dup = df[df_fecha_id.duplicated(keep=False)]
    indices = df_dup[df_dup["campaña"] == 1].index
    return indices



def scale_features(df, columns_to_scale):
    '''
    Función auxiliar que normaliza las columnas indicadas usando StandardScaler

    :param df: dataframe completo
    :param columns_to_scale: lista con nombres de las columnas a normalizar
    :return:
        - dataframe con columnas normalizadas
        - escala de normalización
    '''
    scaler = StandardScaler()
    scaler.fit(df[columns_to_scale])
    df = df.copy()
    df[columns_to_scale] = scaler.transform(df[columns_to_scale])

    return df, scaler.scale_[-1]



def normalizar_visitas(df):
    '''
    Esta función normaliza los valores de la serie temporal visitas de cada producto de acuerdo a la media de cada
    periodo, usando StandardScaler()
    A partir del día 25 de enero de 2016 hay un cambio de tendencia significativo en la media de visitas a la web
    Este incremento de media de visitas diarias, no supone un incremento en la media de ventas diaria.
    Por tanto es mejor para el modelo normalizar los datos de cada período por separado.

    :param df: dataframe completo
    :return:
        dataframe completo con visitas normalizadas
    '''

    lim_norm = datetime.strptime("2016-01-25", "%Y-%m-%d")  # Fecha donde comienza el cambio de tendencia de visitas
    df_norm1 = df[df["fecha"] < lim_norm]
    df_norm1, _ = scale_features(df_norm1, ["visitas"])

    df_norm2 = df[(df["fecha"] >= lim_norm)]
    df_norm2, _ = scale_features(df_norm2, ["visitas"])

    df = df_norm1.append(df_norm2, ignore_index=True)

    return df




def _compute_new_features(df, column, features):
    '''
        Función auxiliar para generar nuevas columnas
    '''
    df = df.copy()
    for feat in features:
        ffunc, params = features[feat]
        ignore_param_in_name = len(params) == 1
        for p in params:
            new_column = f"{column}_{feat}_{p}"
            if ignore_param_in_name:
                new_column = new_column.replace(f"_{p}", "")
            df[new_column] = df.groupby("id")[column].transform(lambda x: ffunc(x, p))

    return df


def generate_features_timeseries(df, columns):
    '''
    Esta función añade atributos a cada instancia con información de la serie temporal especificada.
    Obtiene valores como la diferencia, media, mediana o desviación estándar de una ventana de valores especificada
    :param df: dataframecon los datos
    :param columns: nombre de las columnas sobre las que generar nuevas características
    :return:
        Dataframe con nuevas columnas generadas
    '''

    def diff_feat(x, step):
        return x - x.shift(-step)

    def mean_feat(x, window):
        return x.rolling(window, center=True, min_periods=1).mean()

    def median_feat(x, window):
        return x.rolling(window, center=True, min_periods=1).median()

    def std_feat(x, window):
        return x.rolling(window, center=True, min_periods=1).std(ddof=0)


    # Se obtienen características de los valores en la ventana cercana a cada instancia
    features_functions = {
        "diff": (diff_feat, [-1, -7]),  # Diferencia respecto al día anterior y una semana antes
        "mean": (mean_feat, [3, 7]),    # Media de valores de los últimos 3 y 7 días
        "median": (median_feat, [3, 7]),    # Mediana de los valores de los últimos 3 y 7 días
        "std": (std_feat, [3, 7]),      # Desviación estándar de los valores de los últimos 3 y 7 días
    }

    for column in columns:
        df = _compute_new_features(df, column, features_functions)

    return df



def generate_date_features(df, keep_fecha=False):
    '''
     Esta función genera características asociadas a la fecha de cada instancia (día del año y día de la semana one-hot)
    :param df: dataframe completo
    :param keep_fecha: Booleano que indica si se conserva la columna fecha original en el dataframe
    :return:
    '''
    date_time = pd.to_datetime(df["fecha"])

    df["Yearday"] = date_time.dt.dayofyear

    df["Day"] = date_time.dt.weekday
    df = pd.get_dummies(df, columns=["Day"])

    if not keep_fecha:
        df = df.drop(["fecha"], axis=1)

    return df


def get_model_inputs(df, target="unidades_vendidas", train=True):
    '''
    Esta función genera las instancias X,y para train o test
    :param df: dataframe completo
    :param target: nombre de la columna de la variable a predecir
    :param train: Booleano para generar instancias de train o de test
    :return:
        - X: Instancias con todas las características
        - y: Valores de la variable a predecir
        - columns: Nombre de las columnas usadas
        - df_id_fecha: dataframe que contiene solamente las columnas id y fecha (para concatenar la predicción posteriormente)
    '''
    # Eliminar columnas que no se introducen en el entrenamiento/predicción de los modelos (id, fecha, train/test)
    df = df.copy()
    if train:
        df = df[df["train/test"] == "train"]
    else:
        df = df[df["train/test"] == "test"]

    columns_to_remove = ["id", "fecha", "train/test"]
    df_id_fecha = df[["id", "fecha"]].copy()
    for c in columns_to_remove:
        if c in df.columns:
            df = df.drop(c, axis=1)

    y = df.pop(target).values.astype(np.float32)
    X = df.values

    return X, y


def rRMSE_score(y_true, y_pred):
    return np.sqrt((np.sum((y_true - y_pred) ** 2)) / y_true.shape[0]) / np.mean(y_true)


def cf_score(y_true, y_pred):
    return (
        np.sum([1 for i in range(y_true.shape[0]) if y_true[i] < y_pred[i]])
    ) / y_true.shape[0]


def metric(y_true, y_pred):
    rRMSE = rRMSE_score(y_true, y_pred)
    CF = cf_score(y_true, y_pred)
    return 0.7 * rRMSE + (0.3 * (1 - CF))


def get_error_metrics(y, o):
    """
    Esta función calcula diferentes métricas de error de las predicciones.
    """
    metrics = {
        "CUSTOM_METRIC": metric(y, o),
        "rRMSE": rRMSE_score(y, o),
        "CF": cf_score(y, o),
        "MAE": mean_absolute_error(y, o),
        "MSE": mean_squared_error(y, o),
        "Pearson R": pearsonr(y, o)[0],
    }
    return metrics