from prediccion_utils import read_dataset, normalizar_visitas, \
    generate_features_timeseries, get_model_inputs, generate_date_features, \
    generate_normalized_feature
from xgboost import XGBRegressor
from lightgbm import LGBMRegressor
from catboost import CatBoostRegressor
from sklearn.neural_network import MLPRegressor
from sklearn.ensemble import StackingRegressor
import numpy as np
import pandas as pd

##################################################################################
##################################################################################
## Script de predicción de unidades_vendidas
## Las funciones de utilidad se implementan en el fichero prediccion_utils.py
##################################################################################
##################################################################################


#########################################
### 1. CARGA DE DATOS
#########################################
# Lectura de datos que carga el conjunto de train y test en un mismo detaframe con una columna 'train/test'.
# Esto permite poder realizar las transformaciones conjuntamente.
df = read_dataset("../data/Modelar_UH2021.txt", "../data/Estimar2.txt")


#########################################
### 2. TRANSFORMACIÓN DE DATOS
#########################################


# Descartamos columnas poco importantes como categoria_uno, categoria_dos, estado
df = df[["id", "fecha", "visitas", "unidades_vendidas",
         "precio", "antiguedad", "campaña", "dia_atipico",  "train/test"]]


# Normalizar visitas de acuerdo a la media de cada período
df = normalizar_visitas(df)


# Generación de características de la serie temporal visitas para enriquecer la información de cada instancia
df = generate_features_timeseries(df, columns=["visitas"])
df = generate_normalized_feature(df, column="visitas", by="id")
df = generate_normalized_feature(df, column="visitas", by="fecha")

# Generación de características asociadas a la fecha
df = generate_date_features(df, keep_fecha=True)

# Drop NaN de las primeras instancias de train que no tienen pasado
df = df[(df.notna().all(axis=1)) | (df["train/test"] == "test")]



#########################################
### 3. ENTRENAMIENTO DEL MODELO
#########################################

# Transformamos el dataframe en instancias X,y para el entrenamiento
X_train, y_train, col_names, df_id_fecha_train = get_model_inputs(df, train=True)
X_test, _, _, df_id_fecha_test = get_model_inputs(df, train=False)
print(col_names)
print(len(col_names))
# Definición de los 4 modelos que componen el Stacking Ensemble y sus parámetros
models = [
    (
        XGBRegressor,
        {'alpha': 1.057130848628767, 'booster': 'gbtree', 'colsample_bytree': 0.8759072109177561,
          'eta': 0.010827357025784943, 'eval_metric': 'mae', 'gamma': 0.25433890023313477,
          'lambda': 0.03841923030389837, 'max_depth': 11, 'min_child_weight': 0.054737095255489136,
          'n_estimators': 140, 'nthread': 8, 'subsample': 0.8530340998465535}
    ),
    (
        LGBMRegressor,
        {'boosting_type': 'gbdt', 'colsample_bytree': 0.9453147569086666, 
        'learning_rate': 0.010802274632110177, 'max_depth': 15, 'metric': 'mae', 
        'min_child_samples': 43, 'min_child_weight': 0.027007007134653992, 'min_split_gain': 1.0157938047409019e-06, 
        'n_estimators': 180, 'n_jobs': 8, 'num_leaves': 25, 'objective': 'regression', 
        'reg_alpha': 0.011830936061601648, 'reg_lambda': 0.0007179062176282463, 'subsample': 0.30651461916546213, 
        'subsample_freq': 20}
    ),
    (
        CatBoostRegressor,
        {'bagging_temperature': 4.294407776446887, 'bootstrap_type': 'Bayesian', 'depth': 12, 
        'grow_policy': 'Depthwise', 'iterations': 80, 'l2_leaf_reg': 0.001125911335803743, 
        'learning_rate': 0.026927978772766425, 'logging_level': 'Silent', 'loss_function': 'RMSE', 
        'min_data_in_leaf': 3, 'thread_count': 8}
    ),
    (
        XGBRegressor,
        {'alpha': 1.2087791253195683, 'booster': 'gblinear', 'eta': 0.1281473716613697, 
        'eval_metric': 'rmse', 'lambda': 6.955962225286133e-05, 'n_estimators': 110, 'nthread': 8}
    )
]

estimators = [
    (f"{i}", model_fn(**params))
    for i, (model_fn, params) in enumerate(models)
]
# Modelo Stacking
model = StackingRegressor(
    estimators=estimators,
    final_estimator=MLPRegressor(
        hidden_layer_sizes=(32, 16),
        activation="identity",
        batch_size=64,
        solver="adam",
        learning_rate="adaptive",
    ),verbose=1
)

# Entrenamiento
model.fit(X_train, y_train)



#########################################
### 4. GENERACIÓN DE PREDICCIONES
#########################################

o = model.predict(X_test)

# Redondear al entero superior
o = np.ceil(o)


# Transformaciones para generar fichero de predicciones
df_id_fecha_test["UNIDADES"] = o.astype(np.int)
df_id_fecha_test = df_id_fecha_test.rename(columns={"id": "ID", "fecha": "FECHA"})
df_id_fecha_test = df_id_fecha_test[["FECHA", "ID", "UNIDADES"]]
df_id_fecha_test["FECHA"] = df_id_fecha_test["FECHA"].dt.strftime('%d/%m/%Y')
df_id_fecha_test.to_csv("MINERVA_FINAL.txt", index=False, sep="|", encoding='utf-8')